'''Miscellaneous helpers for the ubuntu-repository-cache charm '''

import os
import random
import subprocess
from collections import OrderedDict

# pylint: disable=F0401
from charmhelpers import fetch
from charmhelpers.contrib.hahelpers import cluster
from charmhelpers.core import (
    hookenv,
    host,
    templating,
)

import apache
import squid
import storage
import util
# pylint: enable=F0401

LOG = hookenv.log
SERVICE = 'ubuntu-repository-cache'


def install():
    '''Perform service installation'''

    # Change apt/sources.list to use the main archive URL.
    # This charm may be provide the ubuntu package repository for
    # the cloud where it is being deployed.  In that case the
    # URL for the repository in the cloud image will point to the
    # service provided by this charm and it could be unavailable.
    url = 'http://%s/ubuntu/' % hookenv.config()['sync-host']
    LOG('Changing apt source list url to %s' % url)
    util.change_aptsources_url(url)
    fetch.apt_update(fatal=True)

    storage.setup()
    apache.install()
    squid.install()
    subprocess.check_call(['useradd', '--gid', 'www-data',
                           '--create-home', '--no-user-group',
                           'www-sync'])


def pause():
    '''Temporarily make the charm service unavailable.

    Use this when the metadata does not exist so that users don't
    find an empty mirror.  The easiest way to do this is by stopping
    apache.  This will keep users from accessing the cache while keeping
    squid running.

    Don't stop Squid, because it:
    * takes a long time to stop/start,
    * trashes the in-memory cache when stopped, and
    * can continue serving siblings if left running.
    '''

    apache.stop()


def unpause():
    '''Make the service available after being paused by pause_service()'''

    if not host.service_running('squid-deb-proxy'):
        squid.start()

    apache.start()


def stop():
    '''Stop service and prevent from running in the future'''

    hookenv.close_port(80)
    hookenv.close_port(8000)
    hookenv.close_port(3130)
    hookenv.close_port(4827)
    apache.stop()
    squid.stop()


def start():
    '''Start the service'''

    squid.start()
    apache.start()
    hookenv.open_port(8000)
    hookenv.open_port(3130)
    hookenv.open_port(4827)
    hookenv.open_port(80)


def is_running():
    '''Check if service is running'''

    return apache.service_running() and \
        squid.service_running()


def service_map():
    '''Mapping of configuration files to services.

    This is used by charmhelpers.core.host.restart_on_change to
    restart services when configuration files are altered.
    '''

    return {
        '/etc/squid-deb-proxy/squid-deb-proxy.conf': ['squid-deb-proxy'],
        '/etc/squid-deb-proxy/allowed-networks-src.acl': ['squid-deb-proxy'],
        '/etc/apache2/sites-available/archive_ubuntu_com.conf': ['apache2'],
        }


def reload_on_change(mapping, restart_on_failure=False):
    '''Reload services based on configuration files changing

    (Based on charmhelpers.core.host.restart_on_change)
    This function is used as a decorator, for example::
        @reload_on_change({
            '/etc/ceph/ceph.conf': [ 'cinder-api', 'cinder-volume' ]
            })
        def ceph_client_changed():
            pass  # your code here

    In this example, the cinder-api and cinder-volume services
    would be reloaded if /etc/ceph/ceph.conf is changed by the
    ceph_client_changed function.
    '''
    def wrap(function):
        def wrapped_f(*args):
            checksums = {}
            for path in mapping:
                checksums[path] = host.file_hash(path)
            function(*args)
            restarts = []
            for path in mapping:
                if checksums[path] != host.file_hash(path):
                    restarts += mapping[path]
            services_list = list(OrderedDict.fromkeys(restarts))
            for service_name in services_list:
                LOG('Reloading %s service' % service_name,
                    hookenv.DEBUG)
                host.service_reload(service_name,
                                    restart_on_failure=restart_on_failure)
        return wrapped_f
    return wrap


@reload_on_change(service_map())
def render_configs():
    '''Render the configuration templates for all required services'''

    LOG('Rendering configuration templates')
    config = hookenv.config()

    apache.render_configs()
    squid.render_configs()

    cron_filename = '/etc/cron.d/ubuntu-repository-cache_rsync'
    if cluster.is_elected_leader(None):
        cron_context = {}
        if not config.get('rsync-minutes'):
            config['rsync-minutes'] = random.randint(0, 59)
            config.save()
        cron_context['Minutes'] = config['rsync-minutes']
        cron_context['UnitID'] = hookenv.local_unit()
        cron_context['UnitPath'] = \
            '-'.join(('unit', hookenv.local_unit().replace('/', '-')))
        templating.render('cron/ubuntu-repository-cache_rsync.cron',
                          cron_filename,
                          cron_context)
    else:
        try:
            os.stat(cron_filename)
        except OSError:
            pass
        else:
            os.remove(cron_filename)
