'''Functions related to charm storage maintenance'''

import glob
import os
import subprocess

# pylint can't find the modules  # pylint: disable=F0401
from charmhelpers.contrib.storage.linux import utils as Storage
from charmhelpers.core import (
    fstab,
    hookenv,
    host,
)
# pylint: enable=F0401

LOG = hookenv.log
SERVICE = 'ubuntu-repository-cache'


def mkdir(path, owner='root', group='root', perms=0555, force=False):
    '''Make directories recursively with a umask=0o000.

    This storage module makes calls to mkdir repeatedly with permissions
    that include group write permission.  This would be masked by the
    default 0o022 umask.   This simple helper function will clear
    the umask prior to calling host.mkdir and then restore it when done.
    '''

    prev_umask = os.umask(0o000)
    host.mkdir(path, owner=owner, group=group, perms=perms, force=force)
    os.umask(prev_umask)


def setup_ephemeral(devices=None):
    '''Formats and mounts ephemeral devices

    Each ephemeral device is formated with ext4 and provided a label
    of the format '<service_name>_<#>'

    Devices are mounted at:
     /srv/<service_name>/ephemeral/<#>

    An fstab entry is added for each device to ensure the device is
    mounted on reboot.
    '''

    if not devices:
        LOG('No ephemeral devices provided')
        return

    devno = 0
    base_dir = '/'.join(('/srv', hookenv.service_name()))
    mkdir(base_dir, perms=0o775)
    ephemeral_dir = '/'.join((base_dir, 'ephemeral'))
    mkdir(ephemeral_dir, perms=0o775)
    config = hookenv.config()
    config['ephemeral-mounts'] = []

    for dev in devices.split(','):
        dev = dev.strip()
        LOG('Configuring ephemeral device {}.'.format(dev), hookenv.DEBUG)
        if not Storage.is_block_device(dev):
            LOG('%s is not a block device, skipping' % dev, hookenv.ERROR)
            continue

        partitions = glob.glob(dev + '*').remove(dev) or []
        for part in partitions:
            if Storage.is_device_mounted(part):
                LOG('Ephemeral device had a mounted partition %s skipping' %
                    part, hookenv.ERROR)
                continue

        if Storage.is_device_mounted(dev):
            LOG('Unmounting %s' % dev)
            host.umount(dev)

        label = '_'.join((hookenv.service_name(), str(devno)))
        mountpoint = '/'.join((ephemeral_dir, str(devno)))

        Storage.zap_disk(dev)
        subprocess.check_call(['mkfs', '-t', 'ext4', '-L', label, dev])

        # Remove any existing fstab entry
        ftab = fstab.Fstab()
        entry = ftab.get_entry_by_attr('device', dev)
        if entry:
            ftab.remove_entry(entry)

        # Add to fstab by label, not device name
        mkdir(mountpoint, perms=0o775)
        devname = '='.join(('LABEL', label))
        ftab.add_entry(fstab.Fstab.Entry(devname, mountpoint, 'ext4',
                                         options=None))
        ftab.close()

        subprocess.check_call(['mount', mountpoint])
        os.chmod(mountpoint, 0o775)
        config['ephemeral-mounts'].append(mountpoint)
        devno += 1
    LOG('Ephemeral storage mounted @ {}'.format(config['ephemeral-mounts']),
        hookenv.DEBUG)
    config.save()


def disk_free(path):
    '''Return disk space free, in MB, for the specified path'''

    output = subprocess.check_output(['df', '-BM', path]).split('\n')
    dev, blocks, used, avail, use_percent, mountpoint = output[1].split()
    LOG('disk free for {} is {}MB'.format(path, avail), hookenv.DEBUG)
    return int(avail.rstrip('M'))


def setup():
    '''Set up storage for the charm'''

    config = hookenv.config()
    setup_ephemeral(config['ephemeral-devices'])

    try:
        primary_mount = config['ephemeral-mounts'][0]
        secondary_mounts = config['ephemeral-mounts'][1:]
    except KeyError:
        primary_mount = '/'.join(('/srv', hookenv.service_name()))
        secondary_mounts = []

    # Make file system directories
    LOG('Creating filesystem entries for %s' % SERVICE)
    mkdir(primary_mount, perms=0o775)

    apache_root = config['apache-root'] = '/'.join((primary_mount, 'apache'))
    # Apache serves from www/ and metadata is in data/
    mkdir(apache_root, owner='www-data', group='www-data', perms=0o770)
    mkdir('/'.join((apache_root, 'www')),
               owner='www-data', group='www-data', perms=0o770)
    mkdir('/'.join((apache_root, 'data')),
               owner='www-data', group='www-data', perms=0o770)
    host.symlink('/'.join((apache_root, 'data', 'ubuntu_active')),
                 '/'.join((apache_root, 'www', 'ubuntu')))

    config['squid-disk-caches'] = []
    for mnt in secondary_mounts:
        squid_cache = '/'.join((mnt, 'squid'))
        mkdir(squid_cache, owner='proxy', group='proxy', perms=0o770)
        max_size = disk_free(squid_cache) * 0.95
        if max_size:
            config['squid-disk-caches'].append([squid_cache, max_size])

    # Add the primary mount last for squid.  This allocates it last
    # and saves additional space for metadata mirrors
    squid_cache = '/'.join((primary_mount, 'squid'))
    mkdir(os.path.dirname(squid_cache), owner='proxy', group='proxy',
               perms=0o770)
    mkdir(squid_cache, owner='proxy', group='proxy', perms=0o770)
    max_size = disk_free(squid_cache)
    # Reserve 8GB for metadata and 4GB for the OS
    reserve = 8192 + 4092
    if max_size < reserve:
        LOG('Less than {}MB disk free @ {} for mirror.'.format(reserve,
                                                               primary_mount),
            hookenv.ERROR)
    else:
        max_size = max_size - reserve
        config['squid-disk-caches'].append([squid_cache, max_size])

    config.save()
