'''Miscellaneous helpers for the ubuntu-repository-cache charm '''

import aptsources.sourceslist
import aptsources.distro
import functools
import glob
import os
import pwd
import shutil
import subprocess

# pylint: disable=F0401
from charmhelpers.core import (
    hookenv,
    host,
)
# pylint: enable=F0401

LOG = hookenv.log
SERVICE = 'ubuntu-repository-cache'


def restore_saved_id():
    '''Restore the effective uid/gid from the saved uid/gid'''

    os.seteuid(os.getresuid()[2])
    os.setegid(os.getresgid()[2])


def become_user(uname):
    '''Change the effective uid/gid, $HOME to the user specified by uname

    This also changes the $HOME in the environment.
    '''

    # Change effective uid/gid back to value from saved uid/gid
    restore_saved_id()

    # Change to the requested user's uid/gid and $HOME
    pw_record = pwd.getpwnam(uname)
    os.setegid(pw_record.pw_gid)
    os.seteuid(pw_record.pw_uid)
    os.environ['HOME'] = pw_record.pw_dir


# Submit MP to include this in charmhelpers.core.host
def run_as_user(user_name):
    '''A decorator to run a function as a particular user.

    This works as long as the saved uid (see os.getresuid()[2])
    has permission to change the effective to that of the
    desired uid/gid of the provided user name.

    The $HOME environment variable is also set to the particular user.
    '''
    def decorator(function):
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            # Remember the current effective uid/gid
            orig_id = os.geteuid(), os.getegid()
            orig_home = os.environ.get('HOME')

            # Change to saved uid/gid so that wrapped functions can nest
            # This assumes that the saved id has
            restore_saved_id()

            # Set effective uid/gid for specified user
            pw_record = pwd.getpwnam(user_name)
            os.setegid(pw_record.pw_gid)
            os.seteuid(pw_record.pw_uid)
            os.environ['HOME'] = pw_record.pw_dir

            result = function(*args, **kwargs)

            # Restore effective uid/gid and home
            restore_saved_id()
            os.setegid(orig_id[1])
            os.seteuid(orig_id[0])
            if orig_home:
                os.environ['HOME'] = orig_home
            elif 'HOME' in os.environ:
                del os.environ['HOME']

            return result
        return wrapper
    return decorator


def system_total_mem():
    '''Return the total system memory in megabytes'''

    output = subprocess.check_output(['free', '-m']).split('\n')
    return int(output[1].split()[1])


def change_aptsources_url(url='http://archive.ubuntu.com/ubuntu'):
    '''Change the URL for the current /etc/apt/sources.list

    This affects the URL for all sources in /etc/apt/sources.list
    so care must be taken if the charm has added any custom source
    URLs as it will alter them as well.

    Sources added with charmhelpers.fetch.add_source() are not affected
    as those are added under /etc/apt/sources.list.d/.
    '''

    sourceslist = aptsources.sourceslist.SourcesList()
    sourceslist.refresh()
    if not glob.glob('/etc/apt/sources.list.%s' % SERVICE):
        sourceslist.backup(backup_ext=SERVICE)
    distro = aptsources.distro.get_distro()
    distro.get_sources(sourceslist)
    distro.change_server(url)
    sourceslist.save()


def peer_IPs(reltype='cluster'):
    '''Return a list of IPs for the peers in the specified relation type'''

    ips = []
    for ident in hookenv.relation_ids(reltype=reltype):
        for unit in hookenv.related_units(relid=ident):
            ips.append(hookenv.relation_get(attribute='private-address',
                                            unit=unit, rid=ident))
    return ips


def rsync_should_retry(returncode):
    '''Based on the provided returncode, should rsync be retried or aborted'''

    # Exit values for rsync that will trigger a retry of the operation
    exitval_retry = frozenset((
        5,      # Error starting client-server protocol
        6,      # Daemon unable to append to log-file
        10,     # Error in socket I/O
        11,     # Error in file I/O
        12,     # Error in rsync protocol data stream
        13,     # Errors with program diagnostics
        14,     # Error in IPC code
        20,     # Received SIGUSR1 or SIGINT
        21,     # Some error returned by waitpid()
        22,     # Error allocating core memory buffers
        23,     # Partial transfer due to error
        24,     # Partial transfer due to vanished source files
        25,     # The --max-delete limit stopped deletions
        30,     # Timeout in data send/receive
        35,     # Timeout waiting for daemon connection
    ))

    # Exit values for rsync that will will abort the operation without retry
    exitval_fail = frozenset((
        1,      # Syntax or usage error
        2,      # Protocol incompatibility
        3,      # Errors selecting input/output files, dirs
        4,      # Requested action not supported
    ))

    if returncode in exitval_retry:
        return True

    if returncode in exitval_fail:
        return False

    LOG('rsync command returned unknown exit value %s' % returncode,
        hookenv.DEBUG)
    return False


def clean_metadata_dir(keep=None):
    '''Delete non-active metadata directories'''

    LOG('Cleaning metadata directory')
    apache_data = '/'.join((hookenv.config()['apache-root'], 'data'))
    if not keep:
        active = os.path.realpath('/'.join((apache_data, 'ubuntu_active')))
        canary = '_'.join((active, 'good'))
        keep = set((active, canary))

    glob_path = '/'.join((apache_data, '/ubuntu_[0-9]*'))
    for entry in [os.path.realpath(p) for p in glob.glob(glob_path)]:
        if entry not in keep:
            if os.path.isfile(entry):
                LOG('Removing file {}'.format(entry), hookenv.DEBUG)
                os.remove(entry)
            elif os.path.isdir(entry) and not os.path.islink(entry):
                LOG('Removing tree {}'.format(entry), hookenv.DEBUG)
                shutil.rmtree(entry)


def get_active_metaver():
    '''Get the metadata version of the active metadata

    Returns version or None
    '''

    apache_data = '/'.join((hookenv.config()['apache-root'], 'data'))
    path = os.path.realpath('/'.join((apache_data, 'ubuntu_active')))
    meta_ver = None
    if os.path.exists(path):
        meta_ver = os.path.basename(path)
    return meta_ver


def update_active_symlink(meta_version):
    '''Update the symlink to the active metadata to point to the
    metadata for meta_version.'''

    apache_data = '/'.join((hookenv.config()['apache-root'], 'data'))
    meta_dir = '/'.join((apache_data, meta_version))
    next_link = '/'.join((apache_data, 'ubuntu_next'))
    active_link = '/'.join((apache_data, 'ubuntu_active'))

    if os.path.exists(next_link):
        os.remove(next_link)
    host.symlink(meta_dir, next_link)
    subprocess.check_call(['mv', '-T', next_link, active_link])


def remote_path_exists(user, host_ip, path, localuser='www-sync'):
    '''Use ssh to determine if a remote path exists.

    Remotely execute the stat(1) command on the remote host.'''

    cmd = ('sudo',
           '-u',
           localuser,
           'ssh',
           '{}@{}'.format(user, host_ip),
           'stat {}'.format(path)
           )
    try:
        subprocess.check_call(cmd)
    except subprocess.CalledProcessError as excp:
        # ssh error
        if excp.returncode is 255:
            raise
        else:  # path does not exist
            return False
    return True
